from .errors import Error
from .controller import UnimonControl
from .version import __version__
from flask import Flask, request, jsonify
import json
import logging

API_BASE = "/api/v1"

DEFAULT_MECH = ""

def api(debug, port, mechanism, controller):
  logger = logging.getLogger("api")
  logger.setLevel(logging.DEBUG)
  DEFAULT_MECH = mechanism

  app = Flask('unimon-ctl')
  app.config["DEBUG"] = debug

  @app.route(API_BASE, methods=['GET'])
  def help():
    logger.debug("📩  || api request for help")
    commands = {
      "version": "get the version of unimon-ctl",
      "domain_id/list": "get a list of all clickos routers on a domain",
      "domain_id/router_id/state": "get the state of a given clickos router"
    }
    body = jsonify(commands)
    return body, 200

  @app.route(API_BASE+"/version", methods=['GET'])
  def version():
    logger.debug("📩  || api request for get version")
    version = {
      "app": "unimon-ctl",
      "version": __version__
    }
    body = jsonify(version)
    return body, 200

  @app.route(API_BASE+"/<int:domain_id>/list", methods=['GET'])
  def list_routers(domain_id):
    logger.debug("📩  || api request for list routers")
    try:
      mechanism = request.args.get("mechanism", DEFAULT_MECH)
      routers = controller.get_router_list(mechanism, domain_id)
      return jsonify(routers), 200
    except Error as e:
      logger.error(e.get_pretty())
      return e.get_json()
    

  logger.debug("💬  || running api")
  app.run(host='0.0.0.0', port=port)


# -*- coding: utf-8 -*-
from .version import __version__
from .errors import Error
from .commech import ComMechanism
from .xenstore import XenStore
import logging

class MechanismError(Error):
  def __init__(self, message):
    super().__init__(
      title="mechanism error",
      message=message
    )

class NoFileError(Error):
  def __init__(self, file_name):
    super().__init__(
      title="file not found",
      message="file with name [{}] could not be opened".format(file_name)
    )

class UnimonControl():

  VERSION = "0.01"
  DEFAULT_COM_MECH="xenstore"

  COMMUNICATION_MECHANISMS = {
    "xenstore": XenStore
  }

  def __init__(self):
    self.logger = logging.getLogger("unimon-control")

### OUTPUT CONTROL

  def __error(self, message):
    """ Formats and outputs errors to the logger """
    self.logger.error("🆘  || {}".format(message))
    return None, message

  def messgage(self, message, data=None, logger=None):
    message = "💬  || {} || data: {}".format(message, data)
    if not logger:
      logger = self.logger
    logger.debug(message)
    return data, None

###

### PUBLIC METHODS
  def set_debug(self):
    self.logger.setLevel(logging.DEBUG)

  def get_version(self):
    """ returns the unimon ctl version """
    return __version__

  def get_router_list(self, mechanism, domain_id):
    """ gets the current state of a clickos router """
    mechanism = self.__validate_mechanism(mechanism, domain_id)
    routers = mechanism.get_router_list(domain_id)
    return routers

  def get_router_state(self, mechanism, domain_id, router_id):
    """ gets the current state of a clickos router """
    mechanism = self.__validate_mechanism(mechanism, domain_id)
    router_state = mechanism.get_router_state(domain_id, router_id)
    return router_state

  def install_from_file(self, mechanism, domain_id, config_path, config_name):
    """ install a click config to a clickos doamin from a .click file """
    mechanism = self.__validate_mechanism(mechanism, domain_id)
    config = ""
    try:
      with open(config_path, 'r') as config_file:
        config = config_file.read()
    except:
      raise NoFileError(config_path)
    return mechanism.install_click_config(domain_id, config, config_name)
  
  def start_router(self, mechanism, domain_id, router_id):
    """ starts a click router on a target clickos domain """
    mechanism = self.__validate_mechanism(mechanism, domain_id)
    rid = mechanism.start_router(domain_id, router_id)
    return rid

  def stop_router(self, mechanism, domain_id, router_id):
    """ stops a click router on a target clickos domain """
    mechanism = self.__validate_mechanism(mechanism, domain_id)
    rid = mechanism.stop_router(domain_id, router_id)
    return rid

  def remove_router(self, mechanism, domain_id, router_id, force):
    mechanism = self.__validate_mechanism(mechanism, domain_id)
    mechanism.remove_router(domain_id, router_id, force)
    return

  def get_elem_handler(self, mechanism, domain_id, router_id, element_name, handler_name):
    mechanism = self.__validate_mechanism(mechanism, domain_id)
    return mechanism.get_element_handler(domain_id, router_id, element_name, handler_name)

### PRIVATE METHODS

  def __validate_mechanism(self, mech_name, dom_id):
    mech_name = mech_name.lower()
    if not mech_name in self.COMMUNICATION_MECHANISMS:
      raise MechanismError("invalid communication mechanism given")
    mech = self.COMMUNICATION_MECHANISMS.get(mech_name)(self.logger.level, mech_name, self.messgage)
    mech.test(dom_id)
    return mech


# -*- coding: utf-8 -*-
from .errors import Error
from .controller import UnimonControl
from .api import api

import argparse
import logging
import os

def get_cli_logger():
  """ ⚠️ only needed for development ⚠️ """
  cli_logger = logging.getLogger("cli")
  cli_logger.setLevel(logging.ERROR)
  if os.getenv("UNIMONCTL_DEV", None):
    cli_logger.setLevel(logging.DEBUG)
  return cli_logger

def validate_domain_id(did):
  if did[0] <= 0:
    logging.fatal("🚨  || domain id must be greater than 0")
    exit(1)
  return did[0]

def main():

  cli_logger = get_cli_logger()

  logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s || %(message)s')
  controller = UnimonControl()

  def run_api(args):
    cli_logger.debug("runing api")
    try:
      api(is_debug, args.get("port"), args.get("mechanism"), controller)
      exit(0)
    except Error as e:
      cli_logger.error(e.get_pretty())
      exit(1)

  def list_routers(args):
    cli_logger.debug("listing routers")
    domain_id = validate_domain_id(args.get("domain_id"))
    try:
      routers = controller.get_router_list(args.get("mechanism"), domain_id)
      print("👍  || router list")
      for rtr, state in routers.items():
        print("➕  || router id: \033[1m{}\033[0m || router state: \033[1m{}\033[0m".format(rtr, state))
      exit(0)
    except Error as e:
      cli_logger.error(e.get_pretty())
      exit(1)

  def check_state(args):
    cli_logger.debug("checking state")
    domain_id = validate_domain_id(args.get("domain_id"))
    try:
      state = controller.get_router_state(args.get("mechanism"), domain_id, args.get("router_id"))
      print("👍  || router state || \033[1m{}\033[0m".format(state))
      exit(0)
    except Error as e:
      cli_logger.error(e.get_pretty())
      exit(1)

  def check_config(args):
    cli_logger.debug("checking config")

  def install_config(args):
    logging.debug("installing config")
    domain_id = validate_domain_id(args.get("domain_id"))
    try:
      name = args.get("config-path") if args.get("name") == None else args.get("name")[0]
      rid = controller.install_from_file(args.get("mechanism"), domain_id, args.get("config-path"), name)
      print("👍  || router has been created || \033[1m{}\033[0m".format(rid))
      if args.get("s"):
        controller.start_router(args.get("mechanism"), domain_id, rid)
        print("👍  || router started || \033[1m{}\033[0m".format(rid))
      exit(0)
    except Error as e:
      cli_logger.error(e.get_pretty())
      exit(1)

  def remove_router(args):
    cli_logger.debug("removing router")
    domain_id = validate_domain_id(args.get("domain_id"))
    try:
      controller.remove_router(args.get("mechanism"), domain_id, args.get("router_id"), args.get("f"))
      print("👍  || router has been removed || \033[1m{}\033[0m".format(args.get("router_id")))
      exit(0)
    except Error as e:
      cli_logger.error(e.get_pretty())
      exit(1)

  def start_router(args):
    cli_logger.debug("starting router")
    domain_id = validate_domain_id(args.get("domain_id"))
    try:
      rid = controller.start_router(args.get("mechanism"), domain_id, args.get("router_id"))
      print("👍  || router started || \033[1m{}\033[0m".format(rid))
      exit(0)
    except Error as e:
      cli_logger.error(e.get_pretty())
      exit(1)

  def stop_router(args):
    cli_logger.debug("stopping router")
    domain_id = validate_domain_id(args.get("domain_id"))
    try:
      rid = controller.stop_router(args.get("mechanism"), domain_id, args.get("router_id"))
      print("👍  || router stopped || \033[1m{}\033[0m".format(rid))
      exit(0)
    except Error as e:
      cli_logger.error(e.get_pretty())
      exit(1)

  def get_handler(args):
    cli_logger.debug("getting handler")
    domain_id = validate_domain_id(args.get("domain_id"))
    try:
      data = controller.get_elem_handler(args.get("mechanism"), domain_id, args.get("router_id"), args.get("element"), args.get("handler"))
      print("👍  || data grabbed || \033[1m{}\033[0m".format(data))
      exit(0)
    except Error as e:
      cli_logger.error(e.get_pretty())
      exit(1)

  def print_version(args=None):
    cli_logger.debug("printing version")
    print("👀  ||| Unimon Control Version ||| \033[1m{}\033[0m".format(controller.get_version()))

  # Args
  parser = argparse.ArgumentParser(description="Control ClickOS Xen Domains")

  # - Top Level
  parser.add_argument('--mechanism', type=str, default=controller.DEFAULT_COM_MECH, help="what mechanism to use for inter-domain communication (default: %(default)s)")
  parser.add_argument('--debug', action='store_true', default=False, help="enable debug level logging")
  parser.add_argument('--version', action='store_true', default=False, help="print the version")
  parser.set_defaults(func=print_version)
  subparsers = parser.add_subparsers(help='sub-command help')
  api_parser = subparsers.add_parser('api', help="run the unimon-ctl api")
  list_parser = subparsers.add_parser('list', help="get a list of all clickos routers present on a domain")
  state_parser = subparsers.add_parser('state', help="get the state of clickos router")
  check_parser = subparsers.add_parser('config', help="get the current config name on a clickos router")
  install_parser = subparsers.add_parser('install', help="install a click config to a clickos xen domain")
  remove_parser = subparsers.add_parser('remove', help="remove click router from clickos domain")
  start_parser = subparsers.add_parser('start', help="start a clickos router")
  stop_parser = subparsers.add_parser('stop', help="stop a clickos router")
  handler_parser = subparsers.add_parser('handler', help="get an elements handler value")

  # -- API
  api_parser.add_argument('--port', type=int, default=8080, help="the port to tun the api over")
  api_parser.set_defaults(func=run_api)

  # -- List
  list_parser.add_argument('domain_id', type=int, nargs=1, default=0, help="the xen domain id running a clickos image")
  list_parser.set_defaults(func=list_routers)

  # -- State
  state_parser.add_argument('domain_id', type=int, nargs=1, default=0, help="the xen domain id running a clickos image")
  state_parser.add_argument('router_id', type=int, help="the id of the target clickos router")
  state_parser.set_defaults(func=check_state)

  # -- Config Check
  check_parser.add_argument('domain_id', type=int, nargs=1, default=0, help="the xen domain id running a clickos image")
  check_parser.add_argument('router_id', type=int, help="the id of the target clickos router")
  check_parser.set_defaults(func=check_config)

  # -- Install
  install_parser.add_argument('domain_id', type=int, nargs=1, default=0, help="the xen domain id running a clickos image")
  install_parser.add_argument('config-path', type=str, help="the path of the .click file to use as the config")
  install_parser.add_argument('--name', type=str, nargs=1, help="add a friendly name for the router config")
  install_parser.add_argument('-s', action='store_true', help="once config installed, start the clickos router")
  install_parser.set_defaults(func=install_config)

  # -- Remove
  remove_parser.add_argument('domain_id', type=int, nargs=1, default=0, help="the xen domain id running a clickos image")
  remove_parser.add_argument('router_id', type=int, help="the id of the target router")
  remove_parser.add_argument('-f', action='store_true', help="if router is still running, it will be stopped")
  remove_parser.set_defaults(func=remove_router)

  # -- Start
  start_parser.add_argument('domain_id', type=int, nargs=1, default=0, help="the xen domain id running a clickos image")
  start_parser.add_argument('router_id', type=int, help="the id of the target clickos router")
  start_parser.set_defaults(func=start_router)

  # -- Stop
  stop_parser.add_argument('domain_id', type=int, nargs=1, default=0, help="the xen domain id running a clickos image")
  stop_parser.add_argument('router_id', type=int, help="the id of the target clickos router")
  stop_parser.set_defaults(func=stop_router)

  # -- Handler
  handler_parser.add_argument('domain_id', type=int, nargs=1, default=0, help="the xen domain id running a clickos image")
  handler_parser.add_argument('router_id', type=int, help="the id of the target clickos router")
  handler_parser.add_argument('element', type=str, help="the name of the target element")
  handler_parser.add_argument('handler', type=str, help="the name of the target handler")
  handler_parser.set_defaults(func=get_handler)

  # Parse
  args = parser.parse_args()
  args_dict = vars(parser.parse_args())
  do_version = args_dict.get("version", False)
  if do_version:
    print_version()
  is_debug = args_dict.get("debug", False)
  if is_debug:
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s || %(message)s')
    controller.set_debug()

  args.func(args_dict)




# -*- coding: utf-8 -*-
from .errors import Error
from .utility import inform
import logging

class ComMechNoOpError(Error):

  def __init__(self, name):
    message = "communication mechanism [{}] does not support this".format(name)
    super().__init__(message=message)
    self.title = "Communication Mechanism Error"

class InvalidDomainError(Error):

  def __init__(self, dom_id):
    super().__init__(
      title="domain not found",
      message="this domain [{}] could not be found".format(dom_id),
      code=400
    )

class NonClickOSDomError(Error):

  def __init__(self, dom_id):
    super().__init__(
      title="domain not clickos",
      message="domain [{}] could not be identified as a clickos domain".format(dom_id),
      code=400
    )

class RouterStillRunningError(Error):

  def __init__(self, dom_id, rid):
    super().__init__(
      title="router already running",
      message="router [{}] is still running on domain [{}]".format(rid, dom_id),
      code=400
    )

class StateChangeError(Error):

  def __init__(self, rid, reason):
    super().__init__(
      title="state change error",
      message="could not set the routers state || reason: {}".format(reason),
      code=400
    )

class ComMechanism():

  ROUTER_STATUS = {
    "unknown": "Unknown",
    "running": "Running",
    "stopped": "Halted",
    "error": "Error"
  }

  def __init__(self, log_level, name, message_func):
    self.name = name
    self.message_func = message_func
    self.logger = logging.getLogger(name)
    self.logger.setLevel(log_level)
    inform(self.logger, "communication mechanism: {}".format(self.name))

  def test(self, dom_id):
    return False

  def get_type_name(self):
    ''' returns the name of the mech type '''
    return name

  def get_router_list(self, dom_id):
    raise ComMechNoOpError(self.name)

  def get_router_state(self, dom_id, router_id):
    raise ComMechNoOpError(self.name)

  def get_next_rid(self, dom_id):
    raise ComMechNoOpError(self.name)

  def install_click_config(self, dom_id, config, name):
    ''' returns a router id on success, -1 on fail '''
    raise ComMechNoOpError(self.name)

  def start_router(self, dom_id, router_id):
    ''' starts a router of a given id '''
    raise ComMechNoOpError(self.name)

  def stop_router(self, dom_id, router_id):
    ''' stops a router of a given id '''
    raise ComMechNoOpError(self.name)

  def remove_router(self, dom_id, router_id, force):
    raise ComMechNoOpError(self.name)

  def get_element_handler(self, dom_id, rid, element_name, handler_name):
    raise ComMechNoOpError(self.name)



# -*- coding: utf-8 -*-
from .commech import ComMechanism, InvalidDomainError, NonClickOSDomError, RouterStillRunningError, StateChangeError
from .utility import inform, success
from .errors import Error
try:
  from pyxs import Client as PyXSClient
  from pyxs import PyXSError, InvalidPath
except:
  print("🆘  || unimon-ctl requires the pyxs package for xenstore communication")
import logging
import os
import time

class XenStoreConnectError(Error):

  def __init__(self, extra=""):
    super().__init__(
      title="xenstore connection failure",
      message="failed to connect to the xenstore || information: {}".format(extra)
    )

class XenStoreWriteError(Error):

  def __init__(self, transaction_id):
    super().__init__(
      title="xenstore write failure",
      message="could not write data to xenstore in transaction [{}]".format(transaction_id)
    )

class XenStoreReadError(Error):

  def __init__(self, path):
    super().__init__(
      title="xenstore read failure",
      message="could not read from xenstore [path: {}]".format(path)
    )

class XenStoreDeleteError(Error):

  def __init__(self, path):
    super().__init__(
      title="xenstore delete failure",
      message="could not delete xenstore [path: {}]".format(path)
    )

class XenStore(ComMechanism):

  SOCKET_ENV = "XS_SOCKET"
  DEFAULT_SOCKET_PATH = "/var/run/xenstored/socket"

  CLICKOS_BASE_PATH = "data/clickos"
  UNIMON_BASE_PATH = "data/unimon"

  ENCODING = "utf-8"
  XS_CHUNK = 512
  MAX_TRIES = 3
  CONTROL_RESPONSE_WAIT = 1

  def __init__(self, log_level, name, message_func):
    super().__init__(log_level, name, message_func)
    self.socket_path = os.getenv(self.SOCKET_ENV, self.DEFAULT_SOCKET_PATH)
    inform(self.logger, "using xenstore socket [{}]".format(self.socket_path))

### Test Connection

  def __test_connection(self):
    try:
      with PyXSClient(unix_socket_path=self.socket_path) as client:
        client.connect()
        return
    except:
      pass
    raise XenStoreConnectError("during initial test")

  def __is_clickos(self, dom_id):
    base_path = self.__get_base_path(dom_id)
    try:
      with PyXSClient(unix_socket_path=self.socket_path) as client:
        if client.exists(base_path.encode(self.ENCODING)) == True:
          return
    except:
      pass
    raise NonClickOSDomError(dom_id)

  def test(self, dom_id):
    self.__test_connection() 
    self.__is_clickos(dom_id)
    success(self.logger, "connection via xenstore made")
    return True

### ----------

### Utils

  def __get_base_path(self, dom_id):
    try:
      with PyXSClient(unix_socket_path=self.socket_path) as client:
        domain_path = client.get_domain_path(dom_id)
        if not domain_path == None:
          return "{}/{}".format(domain_path.decode(self.ENCODING), self.CLICKOS_BASE_PATH)
    except:
      pass
    raise InvalidDomainError(dom_id)

### ----------

### XenStore Usage
  # Each returns none on failure

  def __xs_write(self, dom_id, rid, data):
    base_path = self.__get_base_path(dom_id)
    try:
      with PyXSClient(unix_socket_path=self.socket_path) as client:
        attempts = 0
        tr_id = client.transaction()
        while True:
          attempts += 1
          for p, d in data.items():
            path = "{}/{}/{}".format(base_path, str(rid), p).encode(self.ENCODING)
            client.mkdir(path)
            if len(d) != 0:
              value = d.encode(self.ENCODING)
              client.write(path, value)
          if client.commit():
            success(self.logger, "xenstore transaction [{}] complete".format(tr_id))
            return True
          if attempts >= self.MAX_TRIES:
            client.rollback()
            raise XenStoreWriteError(tr_id)
    except:
      pass
    raise XenStoreConnectError("during write")

  def __xs_read(self, dom_id, rid, path_ext):
    base_path = self.__get_base_path(dom_id)
    path = "{}/{}/{}".format(base_path, str(rid), path_ext).encode(self.ENCODING)
    try:
      with PyXSClient(unix_socket_path=self.socket_path) as client:
        value = client.read(path)
        if value:
          return value.decode(self.ENCODING)
    except:
      pass
    raise XenStoreReadError(path.decode(self.ENCODING))

  def __xs_remove_path(self, dom_id, path_ext):
    base_path = self.__get_base_path(dom_id)
    path = "{}/{}".format(base_path, path_ext).encode(self.ENCODING)
    try:
      with PyXSClient(unix_socket_path=self.socket_path) as client:
        client.delete(path)
        return
    except:
      pass
    raise XenStoreDeleteError(path.decode(self.ENCODING))

  def __xs_list_routers(self, dom_id):
    base_path = self.__get_base_path(dom_id)
    path = base_path.encode(self.ENCODING)
    try:
      with PyXSClient(unix_socket_path=self.socket_path) as client:
        raw_list = client.list(path)
        return raw_list
    except:
      pass
    raise XenStoreReadError(path.decode(self.ENCODING))

### ----------

### ClickOS Usage

  def __next_router_id(self, dom_id):
    base_path = self.__get_base_path(dom_id)
    raw_list = self.__xs_list_routers(dom_id)
    if len(raw_list) == 0: return -1
    routers = []
    for value in raw_list:
      value = value.decode(self.ENCODING)
      try:
        value = int(value)
        routers.append(value)
      except:
        continue
    return max(routers)

  def __router_list(self, dom_id):
    base_path = self.__get_base_path(dom_id)
    raw_list = self.__xs_list_routers(dom_id)
    routers = {}
    for value in raw_list:
      value = value.decode(self.ENCODING)
      try:
        value = int(value)
      except:
        continue
      routers[value] = self.__xs_read(dom_id, value, "status")
    return routers

  def __get_router_state(self, dom_id, rid):
    return self.__xs_read(dom_id, rid, "status")

  def __install_config(self, dom_id, config, name):
    nrid = self.__next_router_id(dom_id)
    nrid += 1
    data = {}
    data["control"] = ""
    data["elements"] = ""
    data["config_name"] = name
    data["status"] = self.ROUTER_STATUS.get("stopped")
    i = 0
    while True:
      pos = (i * self.XS_CHUNK)
      if pos >= len(config):
        break
      path_ext = "config/{}".format(i)
      chunk = config[pos:pos+self.XS_CHUNK]
      data[path_ext] = chunk
      i += 1
    self.__xs_write(dom_id, nrid, data)
    return nrid

  
  

### ----------

### Public

  def get_router_state(self, dom_id, router_id):
    return self.__get_router_state(dom_id, router_id)

  def install_click_config(self, dom_id, config, name):
    rid = self.__install_config(dom_id, config, name)
    success(self.logger, "click config installed to router [{}]".format(rid))
    return rid

  def start_router(self, dom_id, rid):
    state = self.__get_router_state(dom_id, rid)
    if state == self.ROUTER_STATUS.get("running"):
      raise StateChangeError(rid, "router already running")
    data = {"status": self.ROUTER_STATUS.get("running")}
    self.__xs_write(dom_id, rid, data)
    return rid

  def stop_router(self, dom_id, rid):
    state = self.__get_router_state(dom_id, rid)
    if state == self.ROUTER_STATUS.get("stopped"):
      raise StateChangeError(rid, "router already stopped")
    data = {"status": self.ROUTER_STATUS.get("stopped")}
    self.__xs_write(dom_id, rid, data)
    return rid

  def remove_router(self, dom_id, rid, force=False):
    state = self.__get_router_state(dom_id, rid)
    if state == self.ROUTER_STATUS.get("running"):
      if force:
        data = {"status": self.ROUTER_STATUS.get("stopped")}
        self.__xs_write(dom_id, rid, data)
      else:
        raise RouterStillRunningError(dom_id, rid)
    self.__xs_remove_path(dom_id, str(rid))
    return

  def get_router_list(self, dom_id):
    return self.__router_list(dom_id)

  def get_element_handler(self, dom_id, rid, element_name, handler_name):
    print(element_name)
    path_ext = "control/read/{}/{}".format(element_name, handler_name)
    data = {path_ext: ""}
    self.__xs_write(dom_id, rid, data)
    attempts = 0
    path_ext = "elements/{}/{}".format(element_name, handler_name)
    time.sleep(2)
    data = self.__xs_read(dom_id, rid, element_name)
    return data

### ----------



# -*- coding: utf-8 -*-
import json

'''
TODO:
 - comment
 - check pep8
'''

class Error(Exception):

  EMOJI = '‼️'

  def __init__(self, title="error", message="unimon-ctl encouted an error with no description", code=500):
    super().__init__(message)
    self.title = title
    self.message = message
    self.code = code

  def get_pretty(self):
    message = "{}  || {} || {}".format(self.EMOJI, self.title, self.message)
    return message

  def get_json(self):
    error = {
      "error": self.title,
      "message": self.message
    }
    return json.dumps(error), self.code



from .commech import CommMechanism

import logging

class Router():

  MAX_NAME_LEN = 32

  ROUTER_STATES = [
    "Unknown",
    "Running",
    "Stopped",
    "Error"
  ]

  def __init__(self, name, com_mech, dom_id, config, router_id=None, state=None):
    self.name = name
    self.com_mech = com_mech
    self.dom_id = dom_id if isinstance(dom_id, int) else 0
    self.config = config
    self.state=(state or self.ROUTER_STATES[0])
    self.router_id = router_id
    if not router_id:
      self.router_id = self.com_mech.get_next_rid(self.dom_id)
    if not self.__valid_config():
      logging.error("!! router created with invalid config !!")
      raise Exception
    if not self.com_mech.test(self.dom_id):
      logging.error("!! router communications test failed !!")
      raise Exception
    
    self.print_info()

  def __valid_config(self):
    ''' dont look here, pls '''
    if not isinstance(self.name, str):
      return False
    if len(self.name) > self.MAX_NAME_LEN:
      return False
    if not isinstance(self.com_mech, CommMechanism):
      return False
    if not isinstance(self.dom_id, int):
      return False
    if self.dom_id < 1:
      return False
    if not isinstance(self.router_id, int):
      return False
    if self.router_id < 0:
      return False
    if not isinstance(self.config, str):
      return False
    return True

  def __get_next_id(self):
    rid = self.com_mech.get_next_rid(self.dom_id)
    if rid > -1:
      return rid
    logging.error("failed to get next router id [domain {}]".format(self.dom_id))
    raise Exception

  def print_info(self):
    logging.debug("router | {}".format(self.name))
    if not self.__valid_config():
      logging.debug(" - has invalid config!")
      return
    logging.debug("~~~~~")
    logging.debug("router | {}".format(self.name))
    logging.debug(" - communication mech | {}".format(self.com_mech.get_type_name()))
    logging.debug(" - xen domain id | {}".format(self.dom_id))
    logging.debug(" - clickos router id | {}".format(self.router_id))
    logging.debug(" - router state | {}".format(self.state))
    logging.debug("~~~~~")

  def get_state(self):
    state = self.com_mech.get_router_state(self.dom_id, self.router_id)
    if not state:
      self.state = self.ROUTER_STATES[0]
    else:
      self.state = state
    return self.state




import setuptools

exec(open('control/version.py').read())

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="unimon-ctl",
  keywords="unikernel clickos unimon",
  version=__version__,
  author="Will Fantom",
  author_email="w.fantom@lancaster.ac.uk",
  description='Control ClickOS Domains (and get unimon data)',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/willfantom/unimon-ctl",
  packages=setuptools.find_packages(),
  classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Unix",
    ],
  install_requires=['pyxs'],
  entry_points={
    'console_scripts': [
        'unimon-ctl=control.cli:main',
        'clickos-ctl=control.cli:main'
    ],
  },
  python_requires='>=3.6',
)